"""
PostFit for GANs replicas
"""

import os
import re
import shutil
import lhapdf
import logging
import pathlib
import argparse

from subprocess import PIPE
from subprocess import Popen
from rich.table import Table
from rich.console import Console

from validphys import lhio
from validphys.core import PDF


console = Console()
logger = logging.getLogger(__name__)


def get_lhapdf_dir():
    """`get_lhapdf_dir` retrieves the path where the LHAPDF
    data are located.
    """
    lhapdf_dir = Popen(["lhapdf-config", "--datadir"], stdout=PIPE)
    lhapdf_pathdir, _ = lhapdf_dir.communicate()
    lhapdf_pathdir = lhapdf_pathdir.decode("utf-8")
    lhapdf_pathdir = lhapdf_pathdir.replace("\n", "")
    return lhapdf_pathdir


def create_symlink(source, destintation):
    """`create_symlink` creates a symbolic link bewteen the data
    genereted in the run folder and the LHAPDF `datadir` folder.

    Parameters
    ----------
    source : str
        Path indicating the source file to be linked to.
    destintation : str
        Paht indicating where the linked file is located.
    """
    return os.symlink(source, destintation)


def replace_num_members(info_file, nbprior, totrep):
    """Replace the value of the `NumMembers` key in PDF set .info file
    by the new total number of MC replicas.

    Parameters
    ----------
    info_file : str
        file containing the information of the PDF
    nbprior :
        number of prior MC replicas
    totrep :
        total number of the new MC of replicas
    """
    subst = f"NumMembers: {totrep+1}"
    pattern = f"NumMembers: {nbprior}"
    file_handle = open(info_file, 'r')
    file_string = file_handle.read()
    file_handle.close()
    file_string = (re.sub(pattern, subst, file_string))
    file_handle = open(info_file, 'w')
    file_handle.write(file_string)
    file_handle.close()


def postgans(pdf_name, ntotal_rep, check=False):
    """`postgans` Links the generated PDF grids in the run directory
    to the correct directory where the LHAPDF data are located. This
    is done by first requesting the path to the LHAPDF `datadir` using
    `get_lhapdf_dir` and then creates a folder where the enhanced PDFs
    replicas will be placed. If the latter already exists, then removes
    it. The linking is done in two steps:

        (1) first, link all the prior replicas
        (2) and then link all the generated replicas

    The `.info` file of the new enhanced PDF is just a copy of the prior.
    However, it has to be checked and assured beforehand that this is
    exactly the same as the `.info` file in the generated replicas.

    Parameters
    ----------
    pdf_name : str
        Prior PDF name
    ntotal_rep : int
        Total number of replicas
    check : bool
        Choose to whether or not make a check by importing the
        enhanced PDF.
    """

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("POSTGANS", justify="center", width=40)
    table.add_row("Link the generated pdfs to LHAPDF datadir.")
    console.print(table)

    # Access paths
    lhapdf_dir = get_lhapdf_dir()
    lhapdf_pth = pathlib.Path(lhapdf_dir).absolute()
    prior_path = lhapdf_pth / f"{pdf_name}"
    gnpdf_path = lhapdf_pth / f"{pdf_name}_enhanced"

    # Create pdf folder if non-existing and remove it if
    # already existing and replace by a new one.
    if gnpdf_path.is_dir():
        logger.warning(f"{gnpdf_path} already exists and will be removed.")
        shutil.rmtree(gnpdf_path)
    gnpdf_path.mkdir(exist_ok=True)

    # Get GANs output grid
    gan_folder = f"{pdf_name}_enhanced"
    gans_grids = pathlib.Path().absolute() / f"{gan_folder}" / "nnfit"

    # Count the number of replicas in the prior folder
    nbfiles_prior = os.listdir(prior_path)
    nbreplicas_prior = len(nbfiles_prior) - 1

    pdf_info = f"{pdf_name}.info"
    gdf_info = f"{pdf_name}_enhanced.info"
    prior_info = os.path.join(prior_path, pdf_info)
    gnpdf_info = os.path.join(gnpdf_path, gdf_info)
    # Copy file to LHAPDF datadir
    shutil.copy(prior_info, gnpdf_info)
    # Replace NumMembers entry
    replace_num_members(gnpdf_info, nbreplicas_prior, ntotal_rep)

    # Loop over the replicas
    for rep in range(1, ntotal_rep + 1):
        gnpdf = f"{pdf_name}_enhanced_{rep:04}.dat"
        gnpdf_file = os.path.join(gnpdf_path, gnpdf)
        gans_file = gans_grids / f"replica_{rep}"
        gans_file = os.path.join(gans_file, f"{gan_folder}.dat")
        create_symlink(gans_file, gnpdf_file)

    # Compute Replica 0
    lhapdf.pathsPrepend(str(lhapdf_pth))
    generated_pdf = PDF(pdf_name + "_enhanced")
    lhio.generate_replica0(generated_pdf)

    if check:
        try:
            console.print("\n• Try loading enhanced Replica 0.", style="bold blue")
            gen_name = pdf_name + "_enhanced"
            lhapdf.mkPDF(gen_name, 0)
        except RuntimeError as exp:
            console.print(
                    f"\n• {pdf_name} might be corrupted according to {exp}.",
                    style="bold blue"
            )
    console.print("\n• Symbolink link added successfully.", style="bold blue")


def arg_parser():
    """Paree input PDF"""
    parser = argparse.ArgumentParser(description="Create set into LHAPDF")
    parser.add_argument("--pdf", help="PDF name", required=True)
    parser.add_argument("--nenhanced", help="Size of enhanced", required=True)
    argument = parser.parse_args()
    return argument


def main():
    args = arg_parser()
    pdf_name = args.pdf
    enhanced_size = int(args.nenhanced)

    postgans(pdf_name, enhanced_size, check=True)


if __name__ == "__main__":
    main()



"""
Module conataning functions that write down the grid from the outputs of
the GANs. This generates the same output file as n3fit.

For futher details, refer to https://github.com/NNPDF/nnpdf/blob/
8eb094f04c73b994502c1cf0f2592f5541e9c284/n3fit/src/n3fit/io/writer.py
"""

import os
from reportengine.compat import yaml


class WriterWrapper:
    """Class that writes the generated grid into a file. This has exactly
    the same format as the N3FIT output ins such a way that the results can
    be evolved using evolven3fit.

    Parameters
    ----------
    outputname : str
        name of the output folder
    fake_pdf : np.array(float)
        array of fake datased
    xgrid : np.array(float)
        array of x-grid
    replica_ind : int
        index of the given replica
    qscale : float
        value of the initial scale
    """

    def __init__(self, outputname, fake_pdf, xgrid, replica_ind, qscale):

        self.xgrid = xgrid
        self.qscale = qscale
        self.fake_pdf = fake_pdf
        self.outputname = outputname
        self.replica_index = replica_ind

    def write_data(self, replica_path):
        """Write the data into an `.exportgrid` file using the `storegrid`
        method.

        Parameters
        ----------
        replica_path_set : str
            path where the replicas are stored
        outputname : str
            name of the output folder
        """

        os.makedirs(replica_path, exist_ok=True)

        # export PDF grid to file
        storegrid(
            self.fake_pdf,
            self.xgrid,
            self.qscale,
            self.outputname,
            self.replica_index,
            replica_path
        )


def storegrid(fake_replica, xgrid, qscale, outputname, replica_ind, replica_path):
    """Store the results into a grid in the same format as the output of N3FIT.

    Parameters
    ----------
    fake_replica : np.array
        array of fake dataset
    xgrid : np.array(float)
        array of points defining the x-grid
    qscale : float
        initial scale at which the computation was done
    outputname : str
        name of the output folder
    replica_ind : int
        index of the given replica
    replica_path : str
        path where the replicas are stored
    """

    lha = fake_replica.T
    data = {
        "replica": replica_ind,
        "q20": qscale,
        "xgrid": xgrid.T.tolist(),
        "labels": [
            "TBAR",
            "BBAR",
            "CBAR",
            "SBAR",
            "UBAR",
            "DBAR",
            "GLUON",
            "D",
            "U",
            "S",
            "C",
            "B",
            "T",
            "PHT",
        ],
        "pdfgrid": lha.tolist(),
    }

    with open(f"{replica_path}/{outputname}.exportgrid", "w") as fs:
        yaml.dump(data, fs)



import re
import ast
import yaml
import time
import pickle
import logging

from tensorflow.keras import backend as K

from hyperopt import fmin, tpe, hp
from hyperopt.mongoexp import MongoTrials
from hyperopt import space_eval, STATUS_OK

from ganpdfs.train import GanTrain
from ganpdfs.filetrials import FileTrials

logger = logging.getLogger(__name__)

RE_FUNCTION = re.compile("(?<=hp\.)\w*(?=\()")
RE_ARGS = re.compile("\(.*\)$")


def recursive_dic(hyperdict):
    """Parse netsed dictionary for hyperopt.

    Parameters
    ----------
    hyperdict: dict
        Dictionary containing information on the hyper-parameter tune

    Returns
    -------
    dict:
        Dictionary where the values of the keys are hp.functions
    """

    newhyperdict = {}
    for key, value in hyperdict.items():
        if isinstance(value, dict):
            newhyperdict[key] = recursive_dic(value)
        else:
            fname = RE_FUNCTION.search(value)
            if fname is None:
                raise ValueError(f"No hp.function found in ${key}:{value}")
            fname = fname[0]
            args = RE_ARGS.search(value)
            if args is None:
                raise ValueError(f"No arguments found in ${key}:{value}")
            vals = ast.literal_eval(args[0])
            newhyperdict[key] = getattr(hp, fname)(*vals)
    return newhyperdict


def load_yaml(runcard_file):
    """Load YAML file.

    Parameters
    ----------
    runcard_file : str
        input runcard file
    """

    with open(runcard_file, "r") as stream:
        runcard = yaml.load(stream, Loader=yaml.FullLoader)
    hyperdict = runcard.get("hyperopt", {})
    newhyperdict = recursive_dic(hyperdict)
    runcard.update(newhyperdict)
    return runcard


def run_hyperparameter_scan(func_train, search_space, max_evals, cluster, folder):
    """Take care of fetching all the necessary information from a runcard file
    in order to perfrom the hyperparameter scan.

    Parameters
    ----------
    func_train : ganpdfs.train.GanTrain
        GANs function that takes care of the training
    search_space :
        search_space
    max_evals : int
        total number of evalutions
    cluster : str
        cluster adresses
    folder : str
        folder to store the results
    """

    logger.info("Performing hyperparameter scan.")
    if cluster:
        trials = MongoTrials(cluster, exp_key="exp1")
    else:
        # Use constum trials in order to save the model after each trials.
        # The following will generate a .json file containing the trials.
        trials = FileTrials(folder, parameters=search_space)
    best = fmin(
        func_train,
        search_space,
        algo=tpe.suggest,
        max_evals=max_evals,
        trials=trials
    )
    # Save the overall best model
    best_setup = space_eval(search_space, best)
    logger.info("Best scan setup:")
    #     pprint.pprint(best_setup)
    with open("%s/best-model.yaml" % folder, "w") as wfp:
        yaml.dump(best_setup, wfp, default_flow_style=False)
    log = "%s/hyperopt_log_{}.pickle".format(time.time()) % folder
    with open(log, "wb") as wfp:
        logger.info(f"Saving trials in {log}")
        pickle.dump(trials.trials, wfp)
    return best_setup


def hyper_train(params, xpdf, pdf):
    """Take care of running the hyperparameter scan by designing a
    search space through which the model tries to find the best
    architecture that maximizes the figure of merit.

    Parameters
    ----------
    params : dict
        dictionary that contains all the input parameters on which the
        search space is based
    xpdf : np.array(float)
        array of x-grid
    pdf : np.array(float)
        input/prior pdf
    """

    # Train on Input/True pdf
    xgan_pdfs = GanTrain(xpdf, pdf, params)

    smm_result = xgan_pdfs.train(nb_epochs=params.get("epochs"))
    # TODO: rename "loss" throughout hyperopt
    return {"loss": smm_result, "status": STATUS_OK}



import logging
import json
import pathlib
import shutil
import numpy as np
import matplotlib.pyplot as plt
import NNPDF as nnpath

from tqdm import tqdm
from tqdm import trange
from ganpdfs.utils import smm
from numpy.random import PCG64
from numpy.random import Generator
from rich.console import Console

from tensorflow.keras import Model
from tensorflow.keras.layers import Input

from ganpdfs.model import WGanModel
from ganpdfs.model import DWGanModel
from ganpdfs.utils import interpol
from ganpdfs.utils import axes_width
from ganpdfs.utils import gan_summary
from ganpdfs.custom import get_optimizer
from ganpdfs.utils import latent_sampling
from ganpdfs.writer import WriterWrapper
from ganpdfs.custom import save_ckpt
from ganpdfs.custom import wasserstein_loss
from ganpdfs.custom import load_generator
from ganpdfs.custom import GradientPenalty
from ganpdfs.custom import turn_on_training


console = Console()
logger = logging.getLogger(__name__)
STYLE = "bold blue"


class GanTrain:
    """The `GanTrain` class controls the training of the Generative
    Adversarial Neural Networks. It sets the intrplay between the two
    neural networks: Generator and Discriminator.

    Parameters
    ----------
    xgrid : np.array(float)
        array of x-grid from which the input PDF grid for the GAN was
        computed. Notice that this does not have to be the same as the
        LHAPDF grid.
    pdfs : tuple(np.array(float), np.array(float))
        The tupple contains two differnent PDF grids: one generated with
        the GAN x-grid and one generated with the LHAPDF grid.
    params: dict
        Dictionary containing all the input parameters.
    """

    def __init__(self, xgrid, pdfs, params):
        self.noise = 100
        self.norm, pdf, self.lhaPDFs = pdfs
        self.xgrid = xgrid
        self.params = params
        dim, flsize, xgssize = pdf.shape
        self.batch = dim * params["batch_size"] // 100
        self.hyperopt = params.get("scan")
        self.rndgen = Generator(PCG64(seed=0))
        self.folder = params.get("save_output")
        discparams = params.get("disc_parameters")

        # Prepare latent space
        self.latent_pdf = latent_sampling(
                pdf,
                params.get("tot_replicas"),
                self.rndgen,
                nsx=params.get("gauss_noise", 1e-5)
        )
        params["latent_space"] = self.latent_pdf

        # Define Models architecture
        if params.get("architecture") == "dnn":
            real_shape = Input(shape=(flsize, xgssize))
            genparams = params.get("gen_parameters")
            if genparams.get("structure") == "standard":
                synt_shape = Input(shape=(self.noise,))
            elif genparams.get("structure") == "custom":
                synt_shape = Input(shape=(flsize, xgssize))
            self.pdf = pdf
            self.gan = WGanModel(self.pdf, params)
        elif params.get("architecture") == "dcnn":
            logger.warning("DCNN is not fully operational yet!")
            self.pdf = pdf.reshape(pdf.shape + (1,))
            real_shape = Input(shape=(flsize, xgssize, 1))
            synt_shape = Input(shape=(self.noise,))
            self.gan = DWGanModel(self.pdf, params, self.noise)
        else:
            raise ValueError("Invalid Achitecture!")

        # Initialize Models
        self.generator = self.gan.generator_model()
        self.critic = self.gan.critic_model()
        self.adversarial = self.gan.adversarial_model(
            self.generator, self.critic
        )

        # Make sure the Models are trainable
        turn_on_training(self.critic, self.generator)

        # Define Graph for the combined Discriminator. It takes both
        # real PDF samples and Latent space/noise. The Latent Space
        # (noise) is run through the Generator to generate the fake
        # PDF replicas. Both the real and fake PDFs run through the
        # Discriminator for evaluation.
        # The implementation below includes a "Gradient Penalty" for
        # the discriminator, to consider it during the training the
        # training, just modify the loss weight to 1.0. In this case
        # the weights clipping in the input runcard needs to be relaxed.
        synt_discr = self.generator(synt_shape)
        discr_real = self.critic(real_shape)
        discr_synt = self.critic(synt_discr)

        partial_gp = GradientPenalty(
                self.critic,
                self.batch,
                params.get("architecture"),
                discparams.get("gp_weight", 10)
        )([real_shape, synt_discr])

        self.discriminator = Model(
                inputs=[real_shape, synt_shape],
                outputs=[discr_real, discr_synt, partial_gp]
        )
        inputloss = discparams.get("loss", "wasserstein")
        if inputloss != "wasserstein":
            dloss = self.discparams.get("loss")
        else: dloss = wasserstein_loss
        opt_name = discparams.get("optimizer")
        ds_optmz = get_optimizer(opt_name)
        self.discriminator.compile(
                optimizer=ds_optmz,
                loss=[dloss, dloss, "mse"],
                loss_weights=[1.0, 1.0, discparams.get("gp_loss", 1.0)]
        )

        if not self.hyperopt and not params.get("use_saved_model"):
            gan_summary(self.critic, self.generator, self.adversarial)

        # Save Checkpoints
        self.ckpt = save_ckpt(self.generator, self.critic, self.adversarial)

    def real_samples(self, arrpdf, batch):
        """Prepare the real samples. This samples a half-batch of real
        dataset and assign to them target labels (+1) indicating that
        the samples are reals.

        Parameters
        ----------
        batch : int
            dimension of the half batch

        Returns
        -------
        tuple(np.array, np.array)
            containing the random real samples and the target
        labels
        """

        #############################################################
        # Training Description:                                     #
        # --------------------                                      #
        # true_pdfs|==> Critic_Model ==>predicted_labels|+1|==>LOSS #
        #                     ^_________________________________|   #
        #                          TUNING / BACKPROPAGATION         #
        #############################################################
        pdf_index = self.rndgen.integers(0, self.pdf.shape[0], batch)
        pdf_batch = self.pdf[pdf_index]
        y_disc = np.ones((batch, 1))
        return pdf_batch, y_disc

    def sample_latent_space(self, batch):
        """Latent vector that is given as input to the generator. (1) For the
        deep neural network, the latent space vector is constructed as linear
        combinations of the real input PDF with a guassian noise. (2) For the
        deep convolutional network case, the latent vector is just a vector
        of random noise as in the standard GAN/WGAN.

        Parameters
        ----------
        batch : int
            dimension of the half batch

        Returns
        -------
        np.array(float)
            array of random numbers
        """

        genparams = self.params.get("gen_parameters")
        noise_dim = genparams.get("noise_dim", 100)
        if self.params["architecture"]=="dcnn" or genparams["structure"]=="standard":
            latent = self.rndgen.random(noise_dim * batch)
            return latent.reshape(batch, noise_dim)
        elif self.params["architecture"]=="dnn" and genparams["structure"]=="custom":
            pdf_index = self.rndgen.integers(0, self.pdf.shape[0], batch)
            pdf_as_latent = self.latent_pdf[pdf_index]
            expected_shape = (batch, self.pdf.shape[1], self.pdf.shape[2])
            assert pdf_as_latent.shape == expected_shape
            return pdf_as_latent
        else:
            raise ValueError("There is an Error in the architecture!")

    def fake_samples(self, generator, batch):
        """Generate fake samples from the Generator. `fake_samples`
        takes input from the latent space and generate synthetic dataset.
        The synthetic dataset are assigned with target labels (1). The
        output of this is then gets fed to the Critic/Discriminator and
        used to train the later.

        Parameters
        ----------
        generator : ganpdfs.model.WGanModel.generator
            generator neural networks
        batch : int
            dimension of the half batch


        Reuturns
        --------
        tuple(np.array, np.array)
            containing samples from the generated data and the target
            labels
        """

        #############################################################
        # Training Description:                                     #
        # --------------------                                      #
        # fake_pdfs|==> Critic_Model ==>predicted_labels|1|==>LOSS  #
        #                     ^_________________________________|   #
        #                          TUNING / BACKPROPAGATION         #
        #############################################################
        noise = self.sample_latent_space(batch)
        pdf_fake = generator.predict(noise)
        y_disc = -np.ones((batch, 1))
        return pdf_fake, y_disc

    def sample_output_noise(self, batch):
        """Sample the latent space for a given batch and returns the
        corresponding predictions--which is a vector of (-1).

        Parameters
        ----------
        batch: int
            dimension of the batch

        Returns
        -------
        tuple(np.array, np.array)
            noises and the corresponding target labels
        """

        #####################################################################
        # Training Description:                                             #
        # --------------------                                              #
        # noise|==> Generator ==> generated_pdfs|==>Critic==>Label==>LOSS   #
        #              ^______________________________________________|     #
        #                        TUNING / BACKPROPAGATION                   #
        #####################################################################
        noise = self.sample_latent_space(batch)
        y_gen = -np.ones((batch, 1))
        return noise, y_gen

    def dummy_samples(self, batch):
        """Dummy samples that is passed to the gradient penalty. This
        is not used during the training.

        Parameters
        ----------
        batch: int
            dimension of the batch

        Returns
        -------
        np.array(int):
            Vector of zeros
        """

        return np.zeros((batch, 1))

    def train(self, nb_epochs=1000):
        """Train the GANs networks for a given batch size. The training
        is done in such a way that the training of the generator and the
        critic/discriminator is well balanced (more details to be added).

        In order to to be able to evolve the generated grids, the format
        of the x-grid has to be the same as the default x-grid in the
        central replicas file. If this is no the case, then this function
        also performs the interpolation.

        The grids are then written into a file using the `WriterWrapper`
        module.

        In case `hyperopt` is on, the similarity metric--that is used as
        a sort of figrue of merit to assess the performance of the model
        --is computed. This is afterwards used by the 'hyperscan' module.

        Parameters
        ----------
        nb_epochs : int
            total number of epochs

        Returns
        -------
        float:
            similarity metric value
        """

        metric = 0  # Initialize the value of metric
        if not self.params.get("use_saved_model"):
            rdloss, fdloss, advloss = [], [], []
            with trange(nb_epochs, disable=self.hyperopt) as iter_range:
                for k in iter_range:
                    iter_range.set_description("Training")
                    for _ in range(self.params.get("nd_steps", 4)):
                        r_input, r_ydisc = self.real_samples(
                                self.pdf, self.batch
                        )
                        f_input, f_ydisc = self.sample_output_noise(
                                self.batch
                        )
                        d_ydisc = self.dummy_samples(self.batch)
                        dloss = self.discriminator.train_on_batch(
                                [r_input, f_input],
                                [r_ydisc, f_ydisc, d_ydisc]
                        )
                    for _ in range(self.params.get("ng_steps", 3)):
                        noise, y_gen = self.sample_output_noise(self.batch)
                        gloss = self.adversarial.train_on_batch(noise, y_gen)
                    iter_range.set_postfix(rdisc=dloss[0], fdisc=dloss[1], adv=gloss)

                    if not self.hyperopt and (k + 1) % 100 == 0:
                        rdloss.append(dloss)
                        advloss.append(gloss)

            # Save generator model into a folder
            self.generator.save("pre-trained-model")

            if not self.hyperopt:
                loss_info = [{"rdloss": rdloss, "fdloss": fdloss, "advloss": advloss}]
                output_losses = self.params.get("save_output")
                with open(f"{output_losses}/losses_info.json", "w") as outfile:
                    json.dump(loss_info, outfile, indent=2)
        else:
            # Generate fake replicas with the trained model
            console.print(
                "\n• Making predictions using a pre-trained Generator model.",
                style="bold magenta"
            )
            self.generator = load_generator("pre-trained-model")

        fake_pdf, _ = self.fake_samples(self.generator, self.params.get("out_replicas"))

        if not self.hyperopt:

            xgrid = self.xgrid
            # Undo normalization
            fake_pdf = (fake_pdf * self.norm) + self.norm
            #############################################################
            # Interpolate the grids if the GANS-grids is not the same   #
            # as the input PDF.                                         #
            #############################################################
            if self.params.get("architecture") == "dcnn":
                fake_pdf = np.squeeze(fake_pdf, axis=3)
            if self.xgrid.shape != self.params.get("pdfgrid").shape:
                xgrid = self.params.get("pdfgrid")
                console.print("\n• Interpolate GANs grid to LHAPDF grid:", style=STYLE)
                fake_pdf = interpol(fake_pdf, self.xgrid, xgrid, mthd="Intperp1D")
            # Combine the PDFs
            comb_pdf = np.concatenate([self.lhaPDFs, fake_pdf])

            #############################################################
            # Construct the output grids in the same structure as the   #
            # N3FIT outputs. This allows for easy evolution.            #
            #############################################################
            console.print("\n• Write grids to file:", style=STYLE)
            with tqdm(total=comb_pdf.shape[0]) as evolbar:
                for rpid, replica in enumerate(comb_pdf, start=1):
                    grid_path = f"{self.folder}/nnfit/replica_{rpid}"
                    write_grid = WriterWrapper(
                            self.folder,
                            replica,
                            xgrid,
                            rpid,
                            self.params.get("q")
                        )
                    write_grid.write_data(grid_path)
                    evolbar.update(1)
                    evolbar.set_description("Progress")

            #############################################################
            # Copy fit runcard to the enhanced folder                   #
            #############################################################
            try:
                fitpath = nnpath.get_results_path()
                fitpath = fitpath + f"{self.params['pdf']}/filter.yml"
                shutil.copy(fitpath, f"{self.folder}/filter.yml")
            except IOError as excp:
                console.print(
                        f"[bold red]WARNING: Fit run card for {self.params['pdf']}"
                        f" not found. Put it manually into {self.folder} in order"
                        f" to run evolven3fit. {excp}"
                )
        else:
            # Compute FID inception score
            metric, _ = smm(self.pdf, fake_pdf)

        return metric



import numpy as np
import tensorflow as tf

from tensorflow.keras import models
from tensorflow.train import Checkpoint
from tensorflow.keras import optimizers
from tensorflow.keras import constraints
from tensorflow.keras import initializers
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer
from tensorflow.keras.constraints import Constraint
from tensorflow.keras.initializers import Zeros
from tensorflow.keras.initializers import Identity
from tensorflow.keras.constraints import MinMaxNorm


def wasserstein_loss(y_true, y_pred):
    """Compute the wasserstein loss from the true and predictions.
    The loss function is implemented by multiplying the expected
    label for each sample by the predicted score (element wise),
    then calculating the mean. For more details, refer to:
    https://arxiv.org/abs/1506.05439

    Parameters
    ----------
    y_true : tf.tensor(float)
        list of estimated probabilities for the true samples
    y_pred : tf.tensor(float)
        list of estimated probabilities for the fake samples

    Returns
    -------
    tf.tensor
        Tensor containing the mean of the predictions and the
        true values.
    """
    return K.mean(y_true * y_pred)


def get_optimizer(optimizer):
    """Overwrite keras default parameters for optimizer.

    Parameters
    ----------
    optimizer : dict
        Dictionary containing information on the optimizer.

    Returns
    -------
    tf.keras.optimizers
        Optimizer.
    """

    learning_rate = optimizer.get("learning_rate")
    optimizer_name = optimizer.get("optimizer_name")
    optimizer_class = optimizers.get(optimizer_name)
    optimizer_class.learning_rate = learning_rate
    return optimizer_class


def get_activation(model_params):
    """Extract activation functions from the input parameters.
    This is necessary for advanced activation functions.

    Parameters
    ----------
    model_params : dict
        Dictionary containing information on the Discriminator
        architecture.


    Returns
    -------
    tf.keras.activations
        Activation function.
    """

    if model_params.get("activation") == "elu":
        from tensorflow.keras.layers import ELU
        return ELU(alpha=1.0)
    elif model_params.get("activation") == "relu":
        from tensorflow.keras.layers import ReLU
        return ReLU()
    elif model_params.get("activation") == "leakyrelu":
        from tensorflow.keras.layers import LeakyReLU
        return LeakyReLU(alpha=0.2)
    else: raise ValueError("Activation not available.")


def get_init(kinit_name):
    """Get keras kernel initializers from runcadr."""

    return getattr(initializers, kinit_name)


def save_ckpt(generator, critic, adversarial):
    """Save the training information into a file. This includes but
    not limited to the information on the wieghts and the biases of
    the given network. The GANs model is a combination of three
    different neural networks (generator, critic/discriminator,
    adversarial) and the information on each one of them are saved.

    For more information on the constructor `Checkpoint` from
    the module `tensorflow.train`, refer to
    https://www.tensorflow.org/api_docs/python/tf/train/Checkpoint

    Parameters
    ----------
    generator : ganpdfs.model.WassersteinGanModel.generator
        generator neural network
    critic : ganpdfs.model.WassersteinGanModel.critic
        critic/discriminator neural network
    adversarial : ganpdfs.model.WassersteinGanModel.adversarial
        adversarial neural network

    Returns
    -------
    A load status object, which can be used to make assertions about
    the status of a checkpoint restoration
    """

    checkpoint = Checkpoint(
            critic=critic,
            generator=generator,
            adversarial=adversarial
    )
    return checkpoint


def turn_on_training(critic, generator):
    for layer in critic.layers:
        layer.trainable = True
    for layer in generator.layers:
        layer.trainable = True
    critic.trainable = True
    generator.trainable = True


def load_generator(model_name):
    """Load a saved/trained keras model from a folder.

    Parameters
    ----------
    model_name : tf.keras.Model
        Name of the saved folder.

    Returns
    -------
    tf.keras.Model
        Saved generator model.
    """

    return models.load_model(model_name)


class WeightsClipConstraint(Constraint):
    """Put constraints on the weights of a given layer.

    Parameters
    ----------
    value : float
        Value to which the weights will be bounded on.

    Returns
    -------
    tf.keras.constraints
        Constraint class.
    """

    def __init__(self, value):
        self.value = value

    def __call__(self, weights):
        return K.clip(weights, -self.value, self.value)

    def get_config(self):
        return {"value": self.value}


class GenKinit(initializers.Initializer):
    """Custom kernel initialization."""

    def __init__(self, valmin=-1, valmax=1):
        self.valmin = valmin
        self.valmax = valmax

    def __call__(self, shape, dtype=None):
        initk = tf.random.uniform(
                shape,
                minval=self.valmin,
                maxval=self.valmax,
                dtype=dtype
        )
        return initk

    def get_config(self):
        config = {
                'minval': self.valmin,
                'maxval': self.valmax
        }
        return config


class GradientPenalty(Layer):
    """Calculates the gradient penalty loss for a batch of "averaged" samples.
    In Improved WGANs, the 1-Lipschitz constraint is enforced by adding a term
    to the loss function that penalizes the network if the gradient norm moves
    away from 1. However, it is impossible to evaluate this function at all
    points in the input space. The compromise used in the paper is to choose
    random points on the lines between real and generated samples, and check
    the gradients at these points. Note that it is the gradient w.r.t. the input
    averaged samples, not the weights of the discriminator, that we're penalizing!
    In order to evaluate the gradients, we must first run samples through the
    generator and evaluate the loss. Then we get the gradients of the discri-
    minator w.r.t. the input averaged samples. The l2 norm and penalty can then
    be calculated for this gradient.
    Note that this loss function requires the original averaged samples as input,
    but Keras only supports passing y_true and y_pred to loss functions. To get
    around this, we make a partial() of the function with the average argument,
    and use that for model training.
    """

    def __init__(self, critic, batch, architect, gp_weight, **kwargs):
        self.critic = critic
        self.batch = batch
        self.architect = architect
        self.gp_weight = gp_weight
        super(GradientPenalty, self).__init__(**kwargs)

    def call(self, inputs):
        # Compute Random Weighted Average
        reals, fakes = inputs
        if self.architect == "dnn":
            alpha = K.random_uniform((self.batch, 1, 1))
        elif self.architect == "dcnn":
            alpha = K.random_uniform((self.batch, 1, 1, 1))
        rand_avg = (alpha * reals) + ((1 - alpha) * fakes)

        gradients = K.gradients(
                self.critic(rand_avg),
                [rand_avg]
        )[0]
        gradients_sqr = K.square(gradients)
        gradients_sqr_sum = K.sum(
                gradients_sqr,
                axis=np.arange(1, len(gradients_sqr.shape))
        )
        gradient_l2_norm = K.sqrt(gradients_sqr_sum)
        gradient_penalty = K.square(1 - gradient_l2_norm)
        gradient_penalty *= self.gp_weight
        return K.mean(gradient_penalty)


class AddLatent(Layer):
    """Keras layer that inherates from the keras `Layer` class. This layer
    class basically expands the input latent space tensors to the generator.

    Parameters
    ----------
    latent_vector : np.array
        Input latent vector.
    """

    def __init__(self, latent_vector, **kwargs):
        shape = latent_vector.shape[0]
        index = np.random.choice(shape, shape-1, replace=False)
        self.latent = latent_vector[index]
        super(AddLatent, self).__init__(**kwargs)

    def call(self, inputs):
        result = K.switch(
                K.cast(inputs.shape[0] is not None, tf.bool),
                lambda: inputs + self.latent[:inputs.shape[0]],
                lambda: inputs
        )
        return result


class ExpLatent(Layer):
    """Keras layer that inherates from the keras `Layer` class. This layer
    class basically expands the input latent space tensors to the generator.

    Parameters
    ----------
    output_dim : int
        Size of the output dimension.
    use_bias : bool
        Add or not biases to the output layer.
    """

    def __init__(self, output_dim, use_bias, **kwargs):
        self.use_bias = use_bias
        self.units = output_dim
        self.binitializer = Zeros()
        self.kinitializer = Identity()
        super(ExpLatent, self).__init__(**kwargs)

    def build(self, input_shape):
        self.kernel = self.add_weight(
            name='kernel',
            shape=(input_shape[-1], self.units),
            initializer=initializers.get(self.kinitializer),
            trainable=False
        )
        if self.use_bias:
            self.bias = self.add_weight(
                name='bias',
                shape=(self.units),
                initializer=self.binitializer,
                trainable=True
            )
        else: self.bias = None
        super(ExpLatent, self).build(input_shape)

    def call(self, inputs):
        output = K.dot(inputs, self.kernel)
        if self.use_bias: output = output + self.bias
        return output


class GenDense(Layer):
    """Custom layer that inherates from the keras `Layer` class. This layer
    class  is proper to the `Generator` and takes the input parameters from
    the input runcard which contains all the parameters for the layer.

    Parameters
    ----------
    output_dim : int
        Size of the output dimension.
    dicparams : dict
        Dictionary containing the layer parameters.
    """

    def __init__(self, output_dim, dicparams, **kwargs):
        const = MinMaxNorm(
                min_value=-dicparams.get("MinNorm", 1),
                max_value=dicparams.get("MaxNorm", 1e-5),
                rate=dicparams.get("NormRate", 1.0),
                axis=0
        )
        bias_init = dicparams.get("bias_initializer", "zeros")
        self.units = output_dim
        self.kconstraint = constraints.get(const)
        self.kinitializer1 = Identity()
        self.kinitializer2 = initializers.get(GenKinit())
        self.binitializer = get_init(bias_init)
        self.use_bias = dicparams.get("use_bias", False)
        self.activation = dicparams.get("activation")
        super(GenDense, self).__init__(**kwargs)

    def build(self, input_shape):
        self.kernel1 = self.add_weight(
            name='kernel',
            shape=(input_shape[-1], self.units),
            initializer=self.kinitializer1,
            trainable=False
        )
        self.kernel2 = self.add_weight(
            name='kernel',
            shape=(input_shape[-1], self.units),
            initializer=self.kinitializer2,
            trainable=True,
            constraint=self.kconstraint
        )
        if self.use_bias:
            self.bias = self.add_weight(
                name='bias',
                shape=(self.units),
                initializer=self.binitializer,
                trainable=True
            )
        else: self.bias = None
        super(GenDense, self).build(input_shape)

    def call(self, inputs):
        output = K.dot(inputs, self.kernel1 + self.kernel2)
        if self.use_bias: output = output + self.bias
        return output


class ExtraDense(Layer):
    """Keras layer that inherates from the keras `Layer` class. This layer
    class takes the input parameters from the input runcard which contains
    all the parameters for the layer.

    Parameters
    ----------
    output_dim : int
        Size of the output dimension.
    dicparams : dict
        Dictionary containing the layer parameters.
    """

    def __init__(self, output_dim, dicparams, **kwargs):
        wc = WeightsClipConstraint(dicparams.get("weights_constraints")) \
                if output_dim != 1 else None
        self.units = output_dim
        self.kconstraint = constraints.get(wc)
        self.binitializer = get_init(dicparams.get("bias_initializer", "zeros"))
        self.kinitializer = get_init(dicparams.get("kernel_initializer"))
        self.use_bias = dicparams.get("use_bias", False)
        self.activation = dicparams.get("activation")
        super(ExtraDense, self).__init__(**kwargs)

    def build(self, input_shape):
        self.kernel = self.add_weight(
            name='kernel',
            shape=(input_shape[-1], self.units),
            initializer=self.kinitializer,
            trainable=True,
            constraint=self.kconstraint
        )
        if self.use_bias:
            self.bias = self.add_weight(
                name='bias',
                shape=(self.units),
                initializer=self.binitializer,
                trainable=True
            )
        else: self.bias = None
        super(ExtraDense, self).build(input_shape)

    def call(self, inputs):
        output = K.dot(inputs, self.kernel)
        if self.use_bias: output = output + self.bias
        return output


class ConvolutePDF(Layer):
    """Convolute the output of the previous layer with
    a subsample of the input/prior replica.

    Parameters
    ----------
    pdf: np.array
        Array of PDF grids
    """

    def __init__(self, pdf, **kwargs):
        self.pdf = K.constant(pdf[0])
        super(ConvolutePDF, self).__init__(**kwargs)

    def compute_output_shape(self, input_shape):
        return (None, self.pdf.shape[1], self.pdf.shape[2])

    def call(self, previous_layer):
        mult = previous_layer * self.pdf
        return mult


class ConvXgrid(Layer):
    """Convolute the output of the previous layer with the input x-grid.
    """

    def __init__(self, output_dim, xval, kinit="glorot_uniform", **kwargs):
        self.units = output_dim
        self.xval = K.constant(xval)
        self.kernel_initializer = initializers.get(kinit)
        super(ConvXgrid, self).__init__(**kwargs)

    def build(self, input_shape):
        self.kernel = self.add_weight(
            name="kernel",
            shape=(K.int_shape(self.xval)[0], input_shape[1], self.units),
            initializer=self.kernel_initializer,
            trainable=True,
        )
        super(ConvXgrid, self).build(input_shape)

    def call(self, x):
        xres = tf.tensordot(x, self.xval, axes=0)
        xfin = tf.tensordot(xres, self.kernel, axes=([1, 2], [0, 1]))
        return xfin

    def compute_output_shape(self, input_shape):
        return (input_shape[0], self.units)


class PreprocessFit(Layer):
    """Add preprocessing to the output of the previous layer. This is
    expected to assure the PDF-like behavior of the generated samples.

    Parameters
    ----------
    xval: np.array
        Array of x-grid
    """

    def __init__(self, xval, trainable=True, kinit="ones", **kwargs):
        self.xval = xval
        self.trainable = trainable
        self.kernel_initializer = initializers.get(kinit)
        super(PreprocessFit, self).__init__(**kwargs)

    def build(self, input_shape):
        self.kernel = self.add_weight(
            name="kernel",
            shape=(2,),
            initializer=self.kernel_initializer,
            trainable=self.trainable,
        )
        super(PreprocessFit, self).build(input_shape)

    def compute_output_shape(self, input_shape):
        return input_shape

    def call(self, pdf):
        xres = self.xval ** self.kernel[0]
        zres = (1 - self.xval) ** self.kernel[1]
        return pdf * xres * zres



import os
import shutil
import logging
import pathlib
import argparse
import numpy as np

# Set tensorflow log level to error only
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

from rich.table import Table
from rich.style import Style
from rich.console import Console

import tensorflow as tf
import tensorflow.python.util.deprecation as deprecation

from ganpdfs.hyperscan import hyper_train
from ganpdfs.pdformat import XNodes
from ganpdfs.pdformat import InputPDFs
from ganpdfs.hyperscan import load_yaml
from ganpdfs.hyperscan import run_hyperparameter_scan

from tensorflow.compat.v1 import ConfigProto
from tensorflow.compat.v1 import InteractiveSession

config = ConfigProto()
config.gpu_options.allow_growth = True
session = InteractiveSession(config=config)

deprecation._PRINT_DEPRECATION_WARNINGS = False

console = Console()
logging.basicConfig(
        level=logging.WARNING,
        format="\033[0;32m[%(levelname)s]\033[97m %(message)s",
    )
logger = logging.getLogger(__name__)

# Random Seeds
tf.random.set_seed(0)


NF = 6      # Number of flavours
Q0 = 1.65   # Initial energy scale


def splash():
    """Splash information."""

    style = Style(color="blue")
    logo = Table(show_header=True, header_style="bold blue", style=style)
    logo.add_column("𝖌𝖆𝖓𝖕𝖉𝖋𝖘", justify="center", width=80)
    logo.add_row("[bold blue]Generative Adversarial Neural Networks (GANs) for PDF replicas.")
    logo.add_row("[bold blue]https://n3pdf.github.io/ganpdfs/")
    logo.add_row("[bold blue]© N3PDF 2021")
    logo.add_row("[bold blue]Authors: Stefano Carrazza, Juan M. Cruz-Martinez, Tanjona R. Rabemananjara")
    console.print(logo)


def normalize_inpdf(input_pdf):
    pdfabsolute = np.absolute(input_pdf)
    normalization = np.max(pdfabsolute) + 1e-8
    normalized_pdf = (input_pdf - normalization) / normalization
    return normalization, normalized_pdf


def posint(value):
    """Checks that a given number is positive."""

    ivalue = int(value)
    if ivalue <= 0:
        raise argparse.ArgumentTypeError(f"Negative values are not allowed: {value}")
    return ivalue


def argument_parser():
    """Parse the input arguments for wganpdfs.
    """
    # read command line arguments
    parser = argparse.ArgumentParser(description="Generate synthetic replicas with GANs.")
    parser.add_argument("-f", "--force", action="store_true")
    parser.add_argument("runcard", help="A json file with the setup.")
    parser.add_argument("-c", "--cluster", help="Enable cluster scan.")
    parser.add_argument("-s", "--hyperopt", type=int, help="Enable hyperopt scan.")
    parser.add_argument("-t", "--totrep", type=posint, help="Number of output replicas.")
    args = parser.parse_args()

    # check the runcard
    if not os.path.isfile(args.runcard):
        raise ValueError("Invalid runcard: not a file.")
    if args.force:
        logger.warning("Running with --force will overwrite existing results.")

    return args


def main():
    """Main controller from which the main parameters are set and defined.
    """
    splash()
    args = argument_parser()

    hps = load_yaml(args.runcard)
    hps["tot_replicas"] = args.totrep
    nf = hps.get("nf", NF)
    qvalue = hps.get("q", Q0)
    hps["save_output"] = str(hps["pdf"]) + "_enhanced"
    out_folder = pathlib.Path(hps["save_output"]).absolute()
    out_folder.mkdir(exist_ok=True)
    shutil.copyfile(args.runcard, f"{out_folder}/input-runcard.json")

    console.print(
            "\n• Computing PDF grids with the following parameters:",
            style="bold blue"
    )
    init_pdf = InputPDFs(hps["pdf"], qvalue, nf)

    # Choose the LHAPDF x-grid by default
    hps["pdfgrid"] = init_pdf.extract_xgrid()
    if hps["x_grid"] == "lhapdf":
        xgrid = hps["pdfgrid"]
    elif hps["x_grid"] == "custom":
        xgrid = XNodes().build_xgrid()
    elif hps["x_grid"] == "standard":
        xgrid = init_pdf.custom_xgrid(nbpoints=hps.get("nb_xpoints", 200))
    else:
        raise ValueError("{hps['x_grid']} is not a valid grid.")

    # Print summary table of PDF grids
    summary = Table(show_header=True, header_style="bold blue")
    summary.add_column("Parameters", justify="left", width=30)
    summary.add_column("Description", justify="center", width=40)
    summary.add_row("Prior PDF set", f"{hps['pdf']}")
    summary.add_row("Input energy Q0", f"{Q0} GeV")
    summary.add_row(
        "x-grid size",
        f"{xgrid.shape[0]} points, x=({xgrid[0]:.4e}, {xgrid[-1]:.4e})"
    )
    console.print(summary)

    pdf = init_pdf.build_pdf(xgrid)
    pdf_lhapdf = init_pdf.lhaPDF_grids()
    norm, npdf = normalize_inpdf(pdf)
    pdfs = (norm, npdf, pdf_lhapdf)

    # Number of Output replicas (Prior+Synthetics)
    hps["input_replicas"] = pdf.shape[0]
    hps["out_replicas"] = hps["input_replicas"] if args.totrep is None else \
            args.totrep - hps["input_replicas"]

    console.print("\n• Training:", style="bold blue")
    # If hyperscan is True
    if args.hyperopt:
        hps["scan"] = True  # Enable hyperscan

        def fn_hyper_train(params):
            return hyper_train(params, xgrid, pdfs)

        # Run hyper scan
        hps = run_hyperparameter_scan(
            fn_hyper_train, hps, args.hyperopt, args.cluster, out_folder
        )

    # Run the best Model and output logs
    hps["scan"] = False
    hps["verbose"] = True
    loss = hyper_train(hps, xgrid, pdfs)



# Some useful Functions

import logging
import numpy as np
from math import gcd
from tqdm import trange
from scipy import interpolate
from scipy.linalg import sqrtm
from rich.console import Console

console = Console()
logger = logging.getLogger(__name__)


def do_nothing(tensor):
    """Function that does nothing.

    Parameters
    ----------
    tensor: tf.tensor
        Tensor
    """
    return tensor


def gan_summary(critic, generator, adversarial):
    console.print("\n• GanPDFs Architecture:", style="bold blue")
    generator.summary()
    critic.summary()
    adversarial.summary()


def axes_width(ax, lw=1):
    """Change width of axes.
    Parameters
    ----------
    ax : matplotlib.axes.Axes
    Figure's axes
    lw : float
    Define width of axes
    """

    for axis in ['top', 'bottom', 'left', 'right']:
        ax.spines[axis].set_linewidth(lw)


def latent_noise(pdf, rndgen, s=0.0):
    """Apply gaussian noise to the latent input.

    Parameters
    ----------
    pdf: np.array
        Input PDF to be noised
    rndgen:
        random number generator
    s: float
        Intensity of the noise
    """
    arrmin = np.min(pdf)
    arrmax = np.max(pdf)
    gaussian = rndgen.normal(loc=0, scale=1, size=pdf.shape)
    reslt = np.clip(pdf * (1 + s * gaussian), arrmin, arrmax)
    return reslt


def latent_sampling(pdf, nb_output, rndgen, nsx=0.0):
    """latent_sampling.

    Parameters
    ----------
    pdf: np.array
        Array/Grid of input PDFs
    nb_output: int
       Total number of replica
    rndgen:
        Random replica generator
    """

    # assert pdf.shape[0] <= nb_output
    if nb_output is None: nbgen = pdf.shape[0]
    else: nbgen = nb_output
    extra_latent = []
    for n in range(nbgen):
        selected = []
        for _ in range(rndgen.integers(1, 4)):
            select = pdf[rndgen.integers(pdf.shape[0])]
            selected.append(select)
        rslt = np.array(selected)
        rslt = np.sum(rslt, axis=0) / rslt.shape[0]
        extra_latent.append(rslt)
    freslt = np.array(extra_latent)
    freslt = latent_noise(freslt, rndgen, s=nsx)
    return freslt


def factorize_number(number):
    """Factorize_number using Pollard's rho algorithm. This takes
    a number a return a list of the factors.

    Example:
        Given an integer 70, this can be factorized a produc of the
        following integers [7,5,2].

    Notice that by definition, 1 is not included.

    Parameters
    ----------
    number : int
        number to be factorized

    Returns
    -------
    list(int)
        list of the factors
    """

    factors = []

    def get_factor(number):
        x = 2
        factor = 1
        x_fixed = 2
        cycle_size = 2

        while factor == 1:
            for count in range(cycle_size):
                if factor > 1:
                    break
                x = (x * x + 1) % number
                factor = gcd(x - x_fixed, number)
            cycle_size *= 2
            x_fixed = x
        return factor

    while number > 1:
        new_number = get_factor(number)
        factors.append(new_number)
        number //= new_number

    return factors


def construct_cnn(number, nb_layer):
    """Factorize_number using Pollard's rho algorithm that is defined
    by the `factorize_number` method. This is used in order to define
    the dimension of the `strides` for performing the convolution in
    the model class `DCNNWassersteinGanModel`.

    The issue is the following: given a pair of two integers (m, n)
    such that n < m, how can we decompose m into n factors.

    Example:
        Given a pair (70, 3), we have [7,5,2]

    If m cannot be decomposed further, then the remaining cases are
    naturally set to 1.

    Example:
        Given a pair (58, 4) we have [29, 2, 1, 1]

    Noe that the final result is sorted.

    Parameters
    ----------
    number : int
        numbker to be factorized
    nb_layer : int
        dimension of the list that contains the factors


    Returns
    -------
    list(int)
        list of the factors with length `nb_layer`
    """

    factorized = factorize_number(number)
    size = len(factorized)
    cnn_dim = sorted(factorized)
    if size < nb_layer:
        for _ in range(1, nb_layer - size + 1):
            cnn_dim.append(1)
    elif size > nb_layer:
        cnn_dim = cnn_dim[:nb_layer - 1]
        new_elem = np.prod(np.array(cnn_dim))
        cnn_dim.append(new_elem)
    else:
        pass
    return list(map(int, cnn_dim))


def interpol(fake_pdf, gan_grid, lhapdf_grid, mthd="Intperp1D"):
    """Interpolate the generated output according to the x-grid in
    order to match with the LHAPDF grid-format. It uses the `interpolate`
    module from `scipy`. For more details, refere to
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html

    Parameters
    ----------
    fake_pdf : np.array
        generated PDF replicas of shape (nb_repl, nb_flv, gan_grid)
    gan_grid : np.array
        custom x-grid of shape (gan_grid,)
    lhapdf_grid :
        lhapdf-like grid of shape (lhapdf_grid,)

    Returns
    -------
    np.array(float)
        fake PDF replica of shape (nb_repl, nb_flv, gan_grid)
    """
    final_grid = []
    with trange(fake_pdf.shape[0]) as iter_range:
        for k in iter_range:
            fl_space = []
            for fl in fake_pdf[k]:
                if mthd == "Intperp1D":
                    f_interpol = interpolate.interp1d(
                            gan_grid,
                            fl,
                            kind="cubic",
                            fill_value="extrapolate"
                    )
                    new_grid = f_interpol(lhapdf_grid)
                elif mthd == "SplineEv":
                    f_interpol = interpolate.splrep(gan_grid, fl, s=0)
                    new_grid = interpolate.splev(lhapdf_grid, f_interpol, der=0)
                elif mthd == "CubicSpline":
                    f_interpol = interpolate.CubicSpline(gan_grid, fl)
                    new_grid = f_interpol(lhapdf_grid)
                elif mthd == "InterpSpline":
                    f_interpol = interpolate.make_interp_spline(gan_grid, fl)
                    new_grid = f_interpol(lhapdf_grid)
                elif mthd == "UnivariateSpline":
                    weights = np.isnan(fl)
                    fl[weights] = 0.
                    f_interpol = interpolate.UnivariateSpline(
                            gan_grid,
                            fl,
                            w=~weights
                    )
                    f_interpol.set_smoothing_factor(0.01)
                    new_grid = f_interpol(lhapdf_grid)
                else:
                    raise ValueError(f"{mthd} is not an interpolation.")
                fl_space.append(new_grid)
                iter_range.set_description(
                        f"{k+1} synthetic replicas out of {fake_pdf.shape[0]}"
                )
            final_grid.append(fl_space)
    return np.array(final_grid)


def smm(prior, generated):
    """Similarity Metric Measure that measures the quality of the
    generated PDF replicas using the `Fréchet Inception Distance`
    (FID).

    TODO: Check how the total/final FIDs is computed.

    Parameters
    ----------
    prior : np.array(float)
        Prior MC PDF replicas of shape (N, Nf, X)
    generated : np.array(float)
        Generated MC PDF replicas of shape (\tilde{N}, Nf, X)

    Returns
    -------
    float:
        FID
    """

    dim = prior.shape
    # Reshape if the architecture is a DCNN
    if prior.ndim > 3:
        prior = prior.reshape((dim[0], dim[1], dim[2]))
        generated = generated.reshape((dim[0], dim[1], dim[2]))

    # Transpose results w.r.t. the flavors
    prior = np.transpose(prior, axes=[1, 0, 2])
    generated = np.transpose(generated, axes=[1, 0, 2])
    # Prepare FID array
    fid_arr = np.zeros(dim[1])

    def compute_fid(fl_prior, fl_generated):
        """Measure the quality of the generated PDF using the `Fréchet
        Inception Distance` (FID). The Frechet distance between two
        multivariate Gaussians X_1 ~ N(mu_1, C_1) and X_2 ~ N(mu_2, C_2)
        is:
            d^2 = ||mu_1 - mu_2||^2 + Tr(C_1 + C_2 - 2 * sqrt(C_1 * C_2)).

        If the generated PDF replica is exactly the same as the prior,
        the value of the FID is zero; that means that the smaller the
        value of the FID is, the similar the generated replica is to
        the prior.

        For details about the FID's Inception Score measure, refer to
        the following: https://arxiv.org/abs/1706.08500

        Parameters
        ----------
        fl_prior : np.array(float)
            Array of prior PDF replica for a given flavor
        fl_generated : np.array(float)
            Array of generated PDF replica for a given flavor

        Returns
        -------
        float:
            FID value
        """

        # calculate mean and covariance statistics
        mu1, sigma1 = fl_prior.mean(axis=0), np.cov(fl_prior, rowvar=False)
        mu2, sigma2 = fl_generated.mean(axis=0), np.cov(fl_generated, rowvar=False)
        # Check if Infs or NaNs and return a big nnumber
        if (np.isnan(mu2).any() or np.isnan(sigma2).any()):
            return np.random.randint(400, 1000)

        # calculate sum squared difference between means
        ssdiff = np.sum((mu1 - mu2) ** 2.0)
        # calculate sqrt of product between cov
        covmean = sqrtm(sigma1.dot(sigma2))
        # check and correct imaginary numbers from sqrt
        if np.iscomplexobj(covmean):
            covmean = covmean.real
        # calculate score
        fid = ssdiff + np.trace(sigma1 + sigma2 - 2.0 * covmean)
        return fid

    for fl in range(dim[1]):
        fid_arr[fl] = compute_fid(prior[fl], generated[fl])
    fids_stdv = np.std(fid_arr)
    fids_mean = np.mean(fid_arr)
    fids_resc = fids_stdv * fids_mean

    return fids_resc, fid_arr



import json
from hyperopt import Trials
from hyperopt import space_eval


def space_eval_trial(space, trial):
    """space_eval_trial.

    Parameters
    ----------
    space :
        space
    trial :
        trial
    """

    for_eval = {}
    for k, v in trial["misc"]["vals"].items():
        if len(v) == 0:
            for_eval[k] = None
        else:
            for_eval[k] = v[0]
    return space_eval(space, for_eval)


class FileTrials(Trials):
    """FileTrials.
    """

    def __init__(
        self,
        replica_path,
        log=None,
        parameters=None,
        exp_key=None,
        refresh=True
    ):
        self._store_trial = False
        self._json_file = "{0}/tries.json".format(replica_path)
        self._parameters = parameters
        if log:
            self.log_info = log.info
        else:
            self.log_info = print
        super(FileTrials, self).__init__(exp_key=exp_key, refresh=refresh)

    def refresh(self):
        """refresh.
        """
        super(FileTrials, self).refresh()

        # write json to disk
        if self._store_trial:
            # self.log_info(f"Storing scan in {self._json_file}")
            local_trials = []
            for idx, t in enumerate(self._dynamic_trials):
                local_trials.append(t)
                local_trials[idx]["misc"]["space_vals"] = space_eval_trial(
                    self._parameters, t
                )

            all_to_str = json.dumps(local_trials, default=str)
            with open(self._json_file, "w") as f:
                f.write(all_to_str)

    def new_trial_ids(self, N):
        """new_trial_ids.

        Parameters
        ----------
        N :
            N
        """

        self._store_trial = False
        return super(FileTrials, self).new_trial_ids(N)

    def new_trial_docs(self, tids, specs, results, miscs):
        """new_trial_docs.

        Parameters
        ----------
        tids :
            tids
        specs :
            specs
        results :
            results
        miscs :
            miscs
        """

        self._store_trial = True
        return super(FileTrials, self).new_trial_docs(tids, specs, results, miscs)



"""
Generating synthetic Monte Carlo (MC) Parton distribution Function (PDF) replicas
using Generative Adversarial Networks (GANs).


This program has been developed within the N3PDF group. (n3pdf.mi.infn.it/)

Documentation:
https://the-hitchhikers-guide-to-packaging.readthedocs.io/en/latest/introduction.html

Authors: - Stefano Carrazza
         - Juan E. Cruz-Martinez
         - Tanjona R. Rabemananjara

License: GPL-3.0, 2020
"""


import pathlib
from setuptools import setup
from setuptools import find_packages

PACKAGE = "ganpdfs"
THIS_DIR = pathlib.Path(__file__).parent
LONG_DESCRIPTION = (THIS_DIR / "README.md").read_text()
REQUIREMENTS = (THIS_DIR / "requirements.txt").read_text()

try:
    import lhapdf
except ImportError:
    print(f"Note: {PACKAGE} requires the installation of LHAPDF")


setup(
    name=PACKAGE,
    version='1.1.0-dev',
    description="GANs for PDF replicas",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="N3PDF",
    author_email="tanjona.rabemananjara@mi.infn.it",
    license="GPL 3.0",
    url="https://github.com/N3PDF/ganpdfs",
    zip_safe=False,
    project_urls={
        "Documentation": "https://n3pdf.github.io/ganpdfs/",
        "Source": "https://github.com/N3PDF/ganpdfs"
    },
    entry_points={
        "console_scripts": [
            "ganpdfs = ganpdfs.scripts.main:main",
            "postgans = ganpdfs.scripts.postgans:main",
        ]
    },
    package_dir={"": "src"},
    packages=find_packages("src"),
    install_requires=REQUIREMENTS,
    classifiers=[
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    setup_requires=["wheel"],
    python_requires='>=3.6'
)



from tensorflow.keras import Model
from tensorflow.keras.layers import Dense
from tensorflow.keras.layers import Input
from tensorflow.keras.layers import Lambda
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Flatten
from tensorflow.keras.layers import Reshape
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Conv2DTranspose
from tensorflow.keras.layers import BatchNormalization

from ganpdfs.utils import do_nothing
from ganpdfs.utils import construct_cnn
from ganpdfs.custom import ExpLatent
from ganpdfs.custom import GenDense
from ganpdfs.custom import ConvolutePDF
from ganpdfs.custom import ExtraDense
from ganpdfs.custom import get_optimizer
from ganpdfs.custom import wasserstein_loss
from ganpdfs.custom import get_activation
from ganpdfs.custom import WeightsClipConstraint


class WGanModel:
    """GAN class that represents the models in terms of Deep Neural
    Networks. It includes 3 Models: the Genator, the Discriminator,
    and the Adversarial Model which controls the training of the
    Generator. Apart from the Adversarial Model, the Generator and
    the Discriminator are not compiled directly.

    Parameters
    ----------
    pdf: np.array(float)
        Input PDF grid
    params: dict
        dictionary containing the input parameters

    """

    def __init__(self, pdf, params):
        self.pdf = pdf
        self.latentvec = params.get("latent_space")
        self.ganparams = params.get("gan_parameters")
        self.discparams = params.get("disc_parameters")
        self.genparams = params.get("gen_parameters")
        _, self.fl_size, self.xg_size = pdf.shape

    def generator_model(self):
        """Generator Model based on a Deep Neural Network. It takes
        as input a latent space vector and gives as output a PDF grid
        sampled on a discrete x-grid.
        """

        gnn_size = self.genparams.get("size_networks")
        noise_dim = self.genparams.get("noise_dim", 100)
        gnn_dim = self.genparams.get("number_nodes")
        gs_activ = get_activation(self.genparams)
        structure = self.genparams.get("structure", "custom")

        # Loop over the Architecture
        if structure == "standard":
            g_input = Input(shape=(noise_dim,))
            g_dense = Lambda(do_nothing)(g_input)
            for it in range(1, gnn_size + 1):
                g_dense = ExtraDense(
                    gnn_dim * (2 ** it),
                    self.genparams
                )(g_dense)
                g_dense = BatchNormalization()(g_dense)
                g_dense = gs_activ(g_dense)
            g_dense = ExtraDense(
                    self.fl_size * self.xg_size,
                    self.genparams
            )(g_dense)
            g_prefl = Reshape((self.fl_size, self.xg_size))(g_dense)
            g_outpt = ConvolutePDF(self.pdf)(g_prefl)
            g_model = Model(g_input, g_outpt, name="Generator")
        elif structure == "custom":
            g_shape = (self.fl_size, self.xg_size,)
            g_input = Input(shape=g_shape)
            g_lambd = Lambda(do_nothing)(g_input)
            g_dense = ExpLatent(self.xg_size, use_bias=False)(g_lambd)
            for it in range(1, gnn_size + 1):
                g_dense = GenDense(self.xg_size, self.genparams)(g_dense)
            g_model = Model(g_input, g_dense, name="Generator")
        assert g_model.output_shape == (None, self.fl_size, self.xg_size)
        return g_model

    def critic_model(self):
        """Discriminator/Critic model based on Deep Neural Networks. It
        takes as input a PDF grid (which could be the real or the synthetic).
        It gives as output logit values.
        """

        dnn_dim = self.discparams.get("number_nodes")
        dnn_size = self.discparams.get("size_networks")
        ds_activ = get_activation(self.discparams)

        # Loop over the Architecture
        d_shape = (self.fl_size, self.xg_size,)
        d_input = Input(shape=d_shape)
        d_hidden = ExtraDense(dnn_dim, self.discparams)(d_input)
        d_hidden = BatchNormalization()(d_hidden)
        d_hidden = ds_activ(d_hidden)
        for it in range(1, dnn_size + 1):
            d_hidden = ExtraDense(
                dnn_dim // (2 ** it),
                self.discparams
            )(d_hidden)
            d_hidden = BatchNormalization()(d_hidden)
            d_hidden = ds_activ(d_hidden)
        d_flatten = Flatten()(d_hidden)
        d_output = ExtraDense(1, self.discparams)(d_flatten)
        d_model = Model(d_input, d_output, name="Discriminator")
        return d_model

    def adversarial_model(self, generator, critic):
        """Adversarial model based on Deep Neural Networks. It controls the
        training of the Generator through the output of the Discriminator.

        Parameters
        ----------
        generator : tf.Model
            generator model
        critic : tf.model
            critic model
        """

        # extract loss & optimizer
        inputloss = self.ganparams.get("loss", "wasserstein")
        if inputloss != "wasserstein":
            advloss = self.ganparams.get("loss")
        else: advloss = wasserstein_loss
        opt_name = self.ganparams.get("optimizer")
        adv_optimizer = get_optimizer(opt_name)

        # freeze Discriminator
        for layer in critic.layers:
            layer.trainable = False

        model = Sequential(name="Adversarial")
        model.add(generator)       # Add Generator Model
        model.add(critic)          # Add Discriminator Model
        model.compile(loss=advloss, optimizer=adv_optimizer)
        return model


class DWGanModel:
    """GAN class that represents the models in terms of Deep Convolu-
    tion Neural Networks. It includes 3 Models: the Generator, the
    Discriminator, and the Adversarial Model which controls the training
    of the Generator. Apart from the Adversarial Model, the Generator
    and the Discriminator are not compiled directly.

    Parameters
    ----------
    pdf: np.array(float)
        Input PDF grid
    params: dict
        dictionary containing the input parameters
    noise_dim: np.array(float)
        vector of random noise
    """

    def __init__(self, pdf, params, noise_dim):
        self.pdf = pdf
        self.params = params
        self.noise_size = noise_dim
        _, self.fl_size, self.xg_size, _ = pdf.shape
        self.ganparams = params.get("gan_parameters")
        self.discparams = params.get("disc_parameters")
        self.genparams = params.get("gen_parameters")

        # Compute DCNN structure
        self.gnn_size = self.genparams.get("size_networks") + 1
        self.dnn_size = self.discparams.get("size_networks") + 1
        self.cnnf = construct_cnn(pdf.shape[1], self.gnn_size)
        self.cnnx = construct_cnn(pdf.shape[2], self.gnn_size)

    def generator_model(self):
        """Generator Model based on Deep Concolutional Neural Networks. It
        takes as input a random noise vector and gives as output a PDF grid
        sampled on a discrete x-grid.
        """

        gnn_dim = self.genparams.get("number_nodes")
        n_nodes = self.cnnf[0] * self.cnnx[0] * gnn_dim
        gs_activ = get_activation(self.genparams)

        g_model = Sequential(name="Generator")
        g_model.add(Dense(n_nodes, input_dim=self.noise_size))
        if self.discparams.get("batchnorm", True):
            g_model.add(BatchNormalization())
        g_model.add(gs_activ)
        g_model.add(Reshape((self.cnnf[0], self.cnnx[0], gnn_dim)))
        # Loop over the number of hidden layers and
        # upSample at every iteration.
        for it in range(1, self.gnn_size):
            g_model.add(
                Conv2DTranspose(
                    gnn_dim // (2 ** it),
                    kernel_size=(4, 4),
                    strides=(self.cnnf[it], self.cnnx[it]),
                    padding="same"
                )
            )
            if self.discparams.get("batchnorm", True):
                g_model.add(BatchNormalization())
            g_model.add(gs_activ)
        g_model.add(
            Conv2DTranspose(
                1,
                kernel_size=(7, 7),
                padding="same"
            )
        )
        return g_model

    def critic_model(self):
        """Discriminator/Critic model based on Deep Convolutional Neural Networks.
        It takes as input a PDF grid (which could be the real or the synthetic).
        It gives as output logit values.
        """

        const = WeightsClipConstraint(0.01)
        dcnn = self.discparams.get("number_nodes")
        ds_activ = get_activation(self.discparams)

        d_input = (self.fl_size, self.xg_size, 1)
        d_model = Sequential(name="Discriminator")
        d_model.add(
            Conv2D(
                dcnn,
                kernel_size=(4, 4),
                strides=(1, 2),
                padding="same",
                kernel_constraint=const,
                input_shape=d_input
            )
        )
        if self.discparams.get("batchnorm", False):
            d_model.add(BatchNormalization())
        d_model.add(ds_activ)
        # Loop over the number of Layers
        # by Downsampling at each iteration
        for it in range(1, self.dnn_size):
            d_model.add(
                Conv2D(
                    dcnn * (2 ** it),
                    kernel_size=(4, 4),
                    strides=(1, 1),
                    padding="same",
                    kernel_constraint=const,
                )
            )
            if self.discparams.get("batchnorm", False):
                d_model.add(BatchNormalization())
            d_model.add(ds_activ)
        # Flatten and output logits
        d_model.add(Flatten())
        d_model.add(Dense(1))
        return d_model

    def adversarial_model(self, generator, critic):
        """Adversarial model based on Deep Convolutional Neural Networks. It
        controls the training of the Generator through the Discriminator.

        Parameters
        ----------
        generator : tf.Model
            generator model
        critic : tf.model
            critic model
        """

        # Call parameters
        if self.ganparams.get("loss") != "wasserstein":
            advloss = self.ganparams.get("loss")
        else: advloss = wasserstein_loss
        opt_name = self.ganparams.get("optimizer")
        adv_optimizer = get_optimizer(opt_name)

        # freeze Discriminator
        for layer in critic.layers:
            layer.trainable = False

        model = Sequential(name="Adversarial")
        model.add(generator)       # Add Generator Model
        model.add(critic)          # Add Discriminator Model
        model.compile(loss=advloss, optimizer=adv_optimizer)
        return model
